#include "sem.h"
#include <errno.h>
#include <pthread.h>
#include <stdatomic.h>
#include <stdio.h>
#include <stdlib.h>

struct SEM
{
  volatile atomic_int counter;
  pthread_mutex_t mutex;
  pthread_cond_t condition;
};

SEM *
semCreate (int initVal)
{
  SEM *sem = malloc (sizeof (SEM));
  if (sem == NULL)
    {
      perror ("malloc");
      return NULL;
    }

  int tmp_errno = pthread_mutex_init (&(sem->mutex), NULL);
  if (tmp_errno != 0)
    {
      errno = tmp_errno;
      perror ("pthread_mutex_init");
      return NULL;
    }

  tmp_errno = pthread_cond_init (&(sem->condition), NULL);
  if (tmp_errno != 0)
    {
      errno = tmp_errno;
      perror ("pthread_cond_init");
      return NULL;
    }

  sem->counter = initVal;
  return sem;
}

void
semDestroy (SEM *sem)
{
  int tmp_errno = pthread_mutex_destroy (&(sem->mutex));
  if (tmp_errno != 0)
    {
      errno = tmp_errno;
      perror ("pthread_mutex_destroy");
    }

  tmp_errno = pthread_cond_destroy (&(sem->condition));
  if (tmp_errno != 0)
    {
      errno = tmp_errno;
      perror ("pthread_mutex_destroy");
    }

  free (sem);
}

void
P (SEM *sem)
{
  int tmp_errno;
  tmp_errno = pthread_mutex_lock (&(sem->mutex));
  if (tmp_errno != 0)
    {
      errno = tmp_errno;
      perror ("pthread_mutex_lock");
      return;
    }

  while (sem->counter <= 0)
    {
      tmp_errno = pthread_cond_wait (&(sem->condition), &(sem->mutex));
      if (tmp_errno != 0)
        {
          errno = tmp_errno;
          perror ("pthread_cond_wait");
          return;
        }
    }
  sem->counter--;

  tmp_errno = pthread_mutex_unlock (&(sem->mutex));
  if (tmp_errno != 0)
    {
      errno = tmp_errno;
      perror ("pthread_mutex_unlock");
      return;
    }
}

void
V (SEM *sem)
{
  sem->counter++;
  int tmp_errno = pthread_cond_signal (&(sem->condition));
  if (tmp_errno != 0)
    {
      errno = tmp_errno;
      perror ("pthread_cond_signal");
    }
}