#include "plist.h"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

static char *path ();
static int callback (pid_t id, const char *input);
static int killCallback (pid_t id, const char *input);
static void backgroundThreads ();

static char *filep = NULL;

int
main (int argc, char **argv)
{
  filep = path ();
  if (filep == NULL)
    {
      exit (EXIT_FAILURE);
    }

  while (1)
    {
      backgroundThreads ();

      fprintf (stdout, "%s: ", filep);
      if (fflush (stdout) == EOF)
        {
          perror ("fflush");
          exit (EXIT_FAILURE);
        }

      char *input = malloc (1338);
      if (input == NULL)
        {
          perror ("malloc");
          exit (EXIT_FAILURE);
        }

      if (fgets (input, 1338, stdin) == NULL)
        {
          if (ferror (stdin))
            {
              perror ("fgets");
              exit (EXIT_FAILURE);
            }
          else
            {
              free (filep);
              free (input);
              walkList (killCallback);
              backgroundThreads ();
              return EXIT_SUCCESS;
            }
        }

      if (input[strlen (input) - 1] != '\n')
        {
          fprintf (stderr, "Zu große Eingabe");
          int c;
          while ((c = fgetc (stdin)) != '\n')
            {
              if (c == EOF)
                {
                  if (ferror (stdin))
                    {
                      perror ("fgetc");
                      exit (EXIT_FAILURE);
                    }
                  else
                    {
                      free (filep);
                      free (input);
                      walkList (killCallback);
                      backgroundThreads ();
                      return EXIT_SUCCESS;
                    }
                }
            }
          free (input);
          continue;
        }

      input[strlen (input) - 1] = '\0';
      char *input2 = strdup (input);
      if (input2 == NULL)
        {
          perror ("strdup");
          exit (EXIT_FAILURE);
        }

      char **input_args = malloc (2 * sizeof (*input_args));
      if (input_args == NULL)
        {
          perror ("malloc");
          exit (EXIT_FAILURE);
        }
      input_args[0] = strtok (input, " \t");
      int input_argc = 1;

      while ((input_args[input_argc] = strtok (NULL, " \t")) != NULL)
        {
          input_args = realloc (input_args,
                                (++input_argc + 1) * sizeof (*input_args));
          if (input_args == NULL)
            {
              perror ("realloc");
              exit (EXIT_FAILURE);
            }
        }

      if (input_args[0] == NULL)
        {
          free (input);
          free (input_args);
          free (input2);
          continue;
        }
      if (strcmp (input_args[0], "cd") == 0)
        {
          if (chdir (input_args[1]) == -1)
            {
              perror ("chdir");
            }
          free (filep);
          filep = path ();
          if (filep == NULL)
            {
              exit (EXIT_FAILURE);
            }
          free (input);
          free (input_args);
          free (input2);
          continue;
        }
      if (strcmp (input_args[0], "jobs") == 0)
        {
          walkList (callback);
          free (input);
          free (input_args);
          free (input2);
          continue;
        }

      pid_t child = fork ();
      if (child == -1)
        {
          perror ("fork");
          exit (EXIT_FAILURE);
        }
      else if (child == 0)
        {
          if (input_args[input_argc - 1][0] == '&')
            { // delete & char
              input_args[input_argc - 1] = 0;
            }
          execvp (input_args[0], input_args);
          perror ("exec");
          exit (EXIT_FAILURE);
        }
      else
        {
          if (input_args[input_argc - 1][0] == '&')
            {
              if (insertElement (child, input2) < 0)
                {
                  fprintf (stderr, "Insert %s failed\n", input2);
                  exit (EXIT_FAILURE);
                }
            }
          else
            {
              int status;
              child = waitpid (child, &status, 0);
              if (child == -1)
                {
                  perror ("waitpid");
                  exit (EXIT_FAILURE);
                }

              fprintf (stdout, "Exitstatus [%s] = %d\n", input2,
                       WEXITSTATUS (status));
              if (fflush (stdout) == EOF)
                {
                  perror ("fflush");
                  exit (EXIT_FAILURE);
                }
            }
          free (input);
          free (input_args);
          free (input2);
        }
    }
}

static char *
path ()
{
  int len = 100;
  char *buf = malloc (len);
  if (buf == NULL)
    {
      perror ("malloc");
      return NULL;
    }

  while (getcwd (buf, len) == NULL)
    {
      if (errno == ERANGE)
        {
          len += 100;
          buf = realloc (buf, len);
          if (buf == NULL)
            {
              perror ("realloc");
              return NULL;
            }
        }
      else
        {
          perror ("getcwd");
          return NULL;
        }
    }

  return buf;
}

static int
callback (pid_t id, const char *input)
{
  fprintf (stdout, "%d: %s\n", id, input);
  if (fflush (stdout) == EOF)
    {
      perror ("fflush");
      return EXIT_FAILURE;
    }
  return EXIT_SUCCESS;
}

static int
killCallback (pid_t id, const char *input)
{
  if (kill (id, SIGKILL) == -1)
    {
      perror ("kill");
      return EXIT_FAILURE;
    }
  return EXIT_SUCCESS;
}

static void
backgroundThreads ()
{
  while (1)
    {
      int status;
      pid_t child = waitpid (-1, &status, WNOHANG);
      if (child == -1)
        {
          if (errno == ECHILD)
            { // no childs
              return;
            }
          else
            {
              perror ("waitpid");
              exit (EXIT_FAILURE);
            }
        }
      else if (child == 0)
        {
          return;
        }
      else
        {
          char *input = malloc (1337);
          if (input == NULL)
            {
              perror ("malloc");
              exit (EXIT_FAILURE);
            }

          if (removeElement (child, input, 1337) == -1)
            {
              fprintf (stderr, "remove failed");
              exit (EXIT_FAILURE);
            }
          if (WIFEXITED (status))
            {
              printf ("exited, status=%d\n", WEXITSTATUS (status));
            }
          else if (WIFSIGNALED (status))
            {
              printf ("killed by signal %d\n", WTERMSIG (status));
            }
          else if (WIFSTOPPED (status))
            {
              printf ("stopped by signal %d\n", WSTOPSIG (status));
            }
          else if (WIFCONTINUED (status))
            {
              printf ("continued\n");
            }

          if (fflush (stdout) == EOF)
            {
              perror ("fflush");
              exit (EXIT_FAILURE);
            }
          free (input);
        }
    }
}
