#include "queue.h"
#include "sem.h"
#include <errno.h>
#include <stdlib.h>

typedef struct elem elem;

struct elem
{
  elem *next;
  char *cmd;
  char *out;
  int flags;
};

static SEM *changes;
static SEM *empty;
static elem *first;

int
queue_init (void)
{
  changes = semCreate (1);
  if (changes == NULL)
    {
      return -1;
    }

  empty = semCreate (0);
  if (empty == NULL)
    {
      semDestroy (changes);
      return -1;
    }

  first = NULL;

  return 0;
}

void
queue_deinit (void)
{
  semDestroy (changes);
  semDestroy (empty);
}

int
queue_put (char *cmd, char *out, int flags)
{
  elem *next = malloc (sizeof (elem));
  if (next == NULL)
    {
      return -1;
    }
  next->next = NULL;
  next->cmd = cmd;
  next->out = out;
  next->flags = flags;

  P (changes);
  if (first == NULL)
    {
      first = next;
    }
  else
    {
      elem *last = first;
      while (last->next != NULL)
        {
          last = last->next;
        }
      last->next = next;
    }
  V (changes);
  V (empty);

  return 0;
}

int
queue_get (char **cmd, char **out, int *flags)
{
  if (cmd == NULL || out == NULL || flags == NULL)
    {
      errno = EINVAL;
      return -1;
    }

  P (empty);
  P (changes);
  if (first == NULL)
    {
      errno = ENOMEDIUM;
      V (changes);
      return -1;
    }
  *cmd = first->cmd;
  *out = first->out;
  *flags = first->flags;
  elem *oldHead = first;
  first = first->next;
  V (changes);
  free (oldHead);
  return 0;
}
