#define _GNU_SOURCE
#include "connection.h"
#include "request.h"
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

int
initConnectionHandler ()
{
  sigset_t sigset;
  sigemptyset (&sigset);
  sigaddset (&sigset, SIGCHLD);

  struct sigaction sig_action
      = { .sa_handler = SIG_IGN, .sa_mask = sigset, .sa_flags = SA_RESTART };

  if (sigaction (SIGCHLD, &sig_action, NULL) == -1)
    {
      perror ("sigaction");
      exit (EXIT_FAILURE);
    }

  return initRequestHandler ();
}

void
handleConnection (int clientSock, int serverSock)
{
  pid_t pid = fork ();
  if (pid == -1)
    {
      perror ("fork");
      close (clientSock);
    }
  else if (pid == 0)
    {
      if (close (serverSock) == -1)
        {
          perror ("close");
          close (clientSock);
          exit (EXIT_FAILURE);
        }

      FILE *rx = fdopen (clientSock, "r");
      if (rx == NULL)
        {
          perror ("fdopen");
          close (clientSock);
          exit (EXIT_FAILURE);
        }

      int clientSock2 = dup (clientSock);
      if (clientSock2 == -1)
        {
          perror ("dup");
          fclose (rx);
          exit (EXIT_FAILURE);
        }

      FILE *tx = fdopen (clientSock2, "r+");
      if (tx == NULL)
        {
          perror ("fdopen");
          fclose (rx);
          close (clientSock2);
          exit (EXIT_FAILURE);
        }

      handleRequest (rx, tx);

      if (fclose (rx) == -1)
        {
          perror ("fclose");
        }
      if (fclose (tx) == -1)
        {
          perror ("fclose");
        }

      exit (EXIT_SUCCESS);
    }
  else
    {
      if (close (clientSock) == -1)
        {
          perror ("close");
        }
    }
}