#include "cmdline.h"
#include "i4httools.h"
#include "request.h"
#include <asm-generic/errno-base.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int
initRequestHandler (void)
{
  return 0;
}

void
handleRequest (FILE *rx, FILE *tx)
{
  int n = 8192;
  char buf[n];
  if (fgets (buf, n, rx) == NULL)
    {
      if (ferror (rx))
        {
          perror ("fgets");
          httpInternalServerError (tx, "");
        }
      else
        {
          httpBadRequest (tx);
          return;
        }
    }

  if (buf[strlen (buf) - 1] != '\n')
    {
      int c;
      while ((c = fgetc (rx)) != '\n')
        {
          if (c == EOF)
            {
              if (ferror (rx))
                {
                  perror ("fgetc");
                  httpInternalServerError (tx, "");
                }
              else
                {
                  httpBadRequest (tx);
                  return;
                }
            }
        }
    }

  char *args[3];
  char *saveptr;
  args[0] = strtok_r (buf, " ", &saveptr);
  if (args[0] == NULL)
    {
      httpBadRequest (tx);
      return;
    }
  args[1] = strtok_r (NULL, " ", &saveptr);
  if (args[1] == NULL)
    {
      httpBadRequest (tx);
      return;
    }
  args[2] = strtok_r (NULL, " ", &saveptr);
  if (args[2] == NULL)
    {
      httpBadRequest (tx);
      return;
    }
  if (strtok_r (NULL, " ", &saveptr) != NULL)
    {
      httpBadRequest (tx);
      return;
    }

  if (strcmp (args[0], "GET") != 0)
    {
      httpBadRequest (tx);
      return;
    }

  if (strcmp (args[2], "HTTP/1.0\r\n") != 0
      && strcmp (args[2], "HTTP/1.0\n") != 0
      && strcmp (args[2], "HTTP/1.1\r\n") != 0
      && strcmp (args[2], "HTTP/1.1\n") != 0)
    {
      httpBadRequest (tx);
      return;
    }

  if (checkPath (args[1]) == -1)
    {
      httpForbidden (tx, args[1]);
      return;
    }

  const char *root = cmdlineGetValueForKey ("wwwpath");
  char path[strlen (root) + strlen (args[1] + 1)];
  strcpy (path, root);
  strcat (path, args[1]);

  FILE *file = fopen (path, "r");
  if (file == NULL)
    {
      perror ("fopen");
      if (errno == EACCES)
        {
          httpForbidden (tx, args[1]);
        }
      else if (errno == ENOENT)
        {
          httpNotFound (tx, args[1]);
        }
      else
        {
          httpInternalServerError (tx, args[1]);
        }
      return;
    }

  httpOK (tx);

  int c;
  while ((c = fgetc (file)) != EOF)
    {
      if (fputc (c, tx) == EOF)
        {
          perror ("fputc");
          return;
        }
    }
  if (ferror (file))
    {
      perror ("fgetc");
      return;
    }

  if (fclose (file) == EOF)
    {
      perror ("fclose");
      return;
    }
}