#include "cmdline.h"
#include "connection.h"
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>

static void
die (char *msg)
{
  perror (msg);
  exit (EXIT_FAILURE);
}

int
main (int argc, char *argv[])
{
  if (argc != 2 && argc != 3)
    {
      fprintf (stderr, "usage: ./sister --wwwpath=<dir> [--port=<p>]\n");
      exit (EXIT_FAILURE);
    }

  printf ("start");

  if (cmdlineInit (argc, argv) == -1)
    {
      die ("cmdlineInit");
    }

  if (initConnectionHandler () == -1)
    {
      fprintf (stderr, "initConnectionHandler failed\n");
      exit (EXIT_FAILURE);
    }

  const char *path = cmdlineGetValueForKey ("wwwpath");
  if (path == NULL)
    {
      fprintf (stderr, "usage: ./sister --wwwpath=<dir> [--port=<p>]\n");
      exit (EXIT_FAILURE);
    }

  const char *port = cmdlineGetValueForKey ("port");
  if (port == NULL)
    {
      port = "2153";
    }

  errno = 0;
  char *endptr;
  long portnr = strtol (port, &endptr, 10);
  if (errno != 0)
    {
      die ("strtol");
    }

  if (port == endptr || *endptr != '\0')
    {
      fprintf (stderr, "wrong port\n");
      exit (EXIT_FAILURE);
    }

  int suckit = socket (AF_INET6, SOCK_STREAM, 0);

  int flag = 1;
  if (setsockopt (suckit, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof (flag))
      == -1)
    {
      die ("setsockopt");
    }

  struct sockaddr_in6 addr = {
    .sin6_family = AF_INET6,     // = AF_INET6
    .sin6_port = htonl (portnr), // Port-Nummer
    .sin6_flowinfo = 0,          // = 0
    .sin6_addr = in6addr_any,    // IPv6-Adresse
    .sin6_scope_id = 0,          // = 0
  };

  if (bind (suckit, (const struct sockaddr *)&addr, sizeof (addr)) == -1)
    {
      die ("bind");
    }

  if (listen (suckit, SOMAXCONN) == -1)
    {
      die ("listen");
    }

  while (1)
    {
      int clientSock = accept (suckit, NULL, NULL);
      if (clientSock == -1)
        {
          continue;
        }
      handleConnection (clientSock, suckit);
    }

  return EXIT_SUCCESS;
}